# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["SolveRevOp"]

from theano import gof
import theano.tensor as tt
from .base_op import CeleriteBaseOp


class SolveRevOp(CeleriteBaseOp):

    func_file = "./solve_rev.cc"
    func_name = "APPLY_SPECIFIC(solve_rev)"

    def make_node(self, *args):
        in_args = [tt.as_tensor_variable(a) for a in args]
        out_args = [a.type() for a in args[:5]]
        return gof.Apply(self, in_args, out_args)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np
import pymc3 as pm

from .sampling import PyMC3Sampler


def test_tuning():
    ndim = 5
    np.random.seed(42)

    sampler = PyMC3Sampler(start=10, window=50, finish=10)
    with pm.Model() as model:
        pm.Normal("x", sd=np.exp(np.random.uniform(-10, 0, ndim)), shape=ndim)

        trace = sampler.tune(tune=200, progressbar=False)

        start, step = sampler._get_start_and_step(start=None, step=None)

        # Expected step size
        expected = []
        for chain in trace._straces.values():
            expected.append(chain.get_sampler_stats("step_size")[-1])
        expected = np.mean(expected)
        assert np.allclose(step.step_size, expected)

        # Make sure that it's not still the default
        assert not np.allclose(step.step_size, 0.25 / model.ndim ** 0.25)



#!/usr/bin/env python

import os
import sys
from setuptools import setup

dirname = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(dirname, "exoplanet"))
from exoplanet_version import __version__  # NOQA


with open(os.path.join(dirname, "requirements.txt"), "r") as f:
    install_requires = f.read().splitlines()


with open(os.path.join(dirname, "README.md"), encoding="utf-8") as f:
    readme = f.read()


setup(
    name="exoplanet",
    version=__version__,
    author="Daniel Foreman-Mackey",
    author_email="foreman.mackey@gmail.com",
    url="https://github.com/dfm/exoplanet",
    license="MIT",
    packages=[
        "exoplanet",
        "exoplanet.gp",
        "exoplanet.orbits",
        "exoplanet.theano_ops",
        "exoplanet.theano_ops.starry",
        "exoplanet.theano_ops.kepler",
        "exoplanet.theano_ops.contact",
        "exoplanet.theano_ops.celerite",
        "exoplanet.theano_ops.interp",
    ],
    description="Fast & scalable MCMC for all your exoplanet needs",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    package_data={"": ["README.md", "LICENSE"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["PyMC3Sampler"]

import logging

import numpy as np

import pymc3 as pm
from pymc3.step_methods.hmc import quadpotential as quad


class PyMC3Sampler(object):
    """A sampling wrapper for PyMC3 with support for dense mass matrices

    This schedule is based on the method used by as described in Section 34.2
    of the `Stan Manual <http://mc-stan.org/users/documentation/>`_.

    All extra keyword arguments are passed to the ``pymc3.sample`` function.

    Args:
        start (int): The number of steps to run as an initial burn-in to find
            the typical set.
        window (int): The length of the first mass matrix tuning phase.
            Subsequent tuning windows will be powers of two times this size.
        finish (int): The number of tuning steps to run to learn the step size
            after tuning the mass matrix.
        dense (bool): Fit for the off-diagonal elements of the mass matrix.

    """

    # Ref: src/stan/mcmc/windowed_adaptation.hpp in stan repo

    def __init__(self, start=75, finish=50, window=25, dense=True, **kwargs):
        self.dense = dense

        self.start = int(start)
        self.finish = int(finish)
        self.window = int(window)
        self.count = 0
        self._current_step = None
        self._current_trace = None
        self.kwargs = kwargs

    def get_step_for_trace(
        self,
        trace=None,
        model=None,
        regular_window=0,
        regular_variance=1e-3,
        **kwargs
    ):
        """Get a PyMC3 NUTS step tuned for a given burn-in trace

        Args:
            trace: The ``MultiTrace`` output from a previous run of
                ``pymc3.sample``.
            regular_window: The weight (in units of number of steps) to use
                when regularizing the mass matrix estimate.
            regular_variance: The amplitude of the regularization for the mass
                matrix. This will be added to the diagonal of the covariance
                matrix with weight given by ``regular_window``.

        """
        model = pm.modelcontext(model)

        # If not given, use the trivial metric
        if trace is None or model.ndim == 1:
            potential = quad.QuadPotentialDiag(np.ones(model.ndim))

        else:
            # Loop over samples and convert to the relevant parameter space;
            # I'm sure that there's an easier way to do this, but I don't know
            # how to make something work in general...
            N = len(trace) * trace.nchains
            samples = np.empty((N, model.ndim))
            i = 0
            for chain in trace._straces.values():
                for p in chain:
                    samples[i] = model.bijection.map(p)
                    i += 1

            if self.dense:
                # Compute the regularized sample covariance
                cov = np.cov(samples, rowvar=0)
                if regular_window > 0:
                    cov = cov * N / (N + regular_window)
                    cov[np.diag_indices_from(cov)] += (
                        regular_variance
                        * regular_window
                        / (N + regular_window)
                    )
                potential = quad.QuadPotentialFull(cov)
            else:
                var = np.var(samples, axis=0)
                if regular_window > 0:
                    var = var * N / (N + regular_window)
                    var += (
                        regular_variance
                        * regular_window
                        / (N + regular_window)
                    )
                potential = quad.QuadPotentialDiag(var)

        return pm.NUTS(potential=potential, **kwargs)

    def _get_sample_kwargs(self, kwargs):
        new_kwargs = dict(kwargs)
        for k, v in self.kwargs.items():
            if k not in new_kwargs:
                new_kwargs[k] = v
        return new_kwargs

    def _extend(self, steps, start=None, step=None, **kwargs):

        kwargs["compute_convergence_checks"] = False
        kwargs["discard_tuned_samples"] = False
        kwargs["draws"] = 2
        kwargs = self._get_sample_kwargs(kwargs)

        # Hide some of the PyMC3 logging
        logger = logging.getLogger("pymc3")
        propagate = logger.propagate
        level = logger.getEffectiveLevel()
        logger.propagate = False
        logger.setLevel(logging.ERROR)

        self._current_trace = pm.sample(
            start=start, tune=steps, step=step, **kwargs
        )
        self.count += steps

        logger.propagate = propagate
        logger.setLevel(level)

    def warmup(self, start=None, step_kwargs=None, **kwargs):
        """Run an initial warmup phase to find the typical set"""
        if step_kwargs is None:
            step_kwargs = {}
        step = self.get_step_for_trace(**step_kwargs)
        self._extend(self.start, start=start, step=step, **kwargs)
        return self._current_trace

    def _get_start_and_step(
        self, start=None, step_kwargs=None, trace=None, step=None
    ):
        if step_kwargs is None:
            step_kwargs = {}
        if trace is not None:
            step = self.get_step_for_trace(trace, **step_kwargs)
        else:
            if trace is None:
                trace = self._current_trace
            if step is None:
                step = self._current_step
        if start is None:
            if trace is not None:
                start = [t[-1] for t in trace._straces.values()]
        return start, step

    def extend_tune(
        self,
        steps,
        start=None,
        step_kwargs=None,
        trace=None,
        step=None,
        **kwargs
    ):
        """Extend the tuning phase by a given number of steps

        After running the sampling, the mass matrix is re-estimated based on
        this run.

        Args:
            steps (int): The number of steps to run.

        """
        if step_kwargs is None:
            step_kwargs = {}
        start, step = self._get_start_and_step(
            start=start, step_kwargs=step_kwargs, trace=trace, step=step
        )
        self._extend(steps, start=start, step=step, **kwargs)
        self._current_step = step
        return self._current_trace

    def tune(self, tune=1000, start=None, step_kwargs=None, **kwargs):
        """Run the full tuning run for the mass matrix

        This will run ``start`` steps of warmup followed by chains with
        exponentially increasing chains to tune the mass matrix.

        Args:
            tune (int): The total number of steps to run.

        """
        model = pm.modelcontext(kwargs.get("model", None))

        ntot = self.start + self.window + self.finish
        if tune < ntot:
            raise ValueError(
                "'tune' must be at least {0}".format(ntot)
                + "(start + window + finish)"
            )

        self.count = 0
        self.warmup(start=start, step_kwargs=step_kwargs, **kwargs)
        steps = self.window
        trace = None
        while self.count < tune:
            trace = self.extend_tune(
                start=start,
                step_kwargs=step_kwargs,
                steps=steps,
                trace=trace,
                **kwargs
            )
            steps *= 2
            if self.count + steps + steps * 2 > tune:
                steps = tune - self.count

        # Final tuning stage for step size
        self.extend_tune(
            start=start,
            step_kwargs=step_kwargs,
            steps=self.finish,
            trace=trace,
            **kwargs
        )

        # Copy across the step size from the parallel runs
        self._current_step.stop_tuning()
        expected = []
        for chain in self._current_trace._straces.values():
            expected.append(chain.get_sampler_stats("step_size")[-1])

        step = self._current_step
        if step_kwargs is None:
            step_kwargs = dict()
        else:
            step_kwargs = dict(step_kwargs)
            step_kwargs.pop("regular_window", None)
            step_kwargs.pop("regular_variance", None)
        step_kwargs["model"] = model
        step_kwargs["step_scale"] = np.mean(expected) * model.ndim ** 0.25
        step_kwargs["adapt_step_size"] = False
        step_kwargs["potential"] = step.potential
        self._current_step = pm.NUTS(**step_kwargs)
        return self._current_trace

    def sample(
        self, trace=None, step=None, start=None, step_kwargs=None, **kwargs
    ):
        """Run the production sampling using the tuned mass matrix

        This is a light wrapper around ``pymc3.sample`` and any arguments used
        there (for example ``draws``) can be used as input to this method too.

        """
        start, step = self._get_start_and_step(
            start=start, step_kwargs=step_kwargs, step=step
        )
        if trace is not None:
            start = None
        kwargs["tune"] = kwargs.get("tune", 0)
        kwargs = self._get_sample_kwargs(kwargs)
        self._current_trace = pm.sample(
            start=start, step=step, trace=trace, **kwargs
        )
        return self._current_trace



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["ContactPointsOp"]

import theano
from theano import gof
import theano.tensor as tt

from ..build_utils import get_cache_version, get_compile_args, get_header_dirs


class ContactPointsOp(gof.COp):
    num_inputs = 7
    params_type = gof.ParamsType(tol=theano.scalar.float64)
    __props__ = ("tol",)
    func_file = "./contact.cc"
    func_name = "APPLY_SPECIFIC(contact)"

    def __init__(self, tol=1e-10, **kwargs):
        self.tol = float(tol)
        super(ContactPointsOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self):
        return get_cache_version()

    def c_compile_args(self, compiler):
        return get_compile_args(compiler)

    def c_headers(self, compiler):
        return ["exoplanet/theano_helpers.h", "exoplanet/contact_points.h"]

    def c_header_dirs(self, compiler):
        return get_header_dirs(eigen=False)

    def make_node(self, *args):
        if len(args) != self.num_inputs:
            raise ValueError("expected {0} inputs".format(self.num_inputs))
        in_args = [tt.as_tensor_variable(a) for a in args]
        out_args = [
            in_args[0].type(),
            in_args[0].type(),
            tt.zeros_like(in_args[0], dtype="int32").type(),
        ]
        return gof.Apply(self, in_args, out_args)

    def infer_shape(self, node, shapes):
        return shapes[0], shapes[0], shapes[0]



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["GetClOp"]

from theano import gof
import theano.tensor as tt

from .base_op import StarryBaseOp
from .get_cl_rev import GetClRevOp


class GetClOp(StarryBaseOp):

    __props__ = ()
    func_file = "./get_cl.cc"
    func_name = "APPLY_SPECIFIC(get_cl)"
    num_input = 1

    def __init__(self):
        self.grad_op = GetClRevOp()
        super(GetClOp, self).__init__()

    def make_node(self, arg):
        return gof.Apply(self, [tt.as_tensor_variable(arg)], [arg.type()])

    def infer_shape(self, node, shapes):
        return (shapes[0],)

    def grad(self, inputs, gradients):
        return (self.grad_op(gradients[0]),)

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["add_citations_to_model", "CITATIONS"]

import logging
import textwrap
import pymc3 as pm


def add_citations_to_model(citations, model=None):
    try:
        model = pm.modelcontext(model)
        if not hasattr(model, "__citations__"):
            model.__citations__ = dict()
        for k in citations:
            model.__citations__[k] = CITATIONS[k]

    except TypeError:
        pass


def get_citations_for_model(model=None, width=79):
    """Get the citations for the components used an exoplanet PyMC3

    Returns: The acknowledgement text for exoplanet and its dependencies and a
    string containing the BibTeX entries for the citations in the
    acknowledgement.

    """
    model = pm.modelcontext(model)
    if not hasattr(model, "__citations__"):
        logging.warning("no citations registered with model")
        return "", ""

    cite = (
        list(CITATIONS["exoplanet"][0])
        + list(CITATIONS["pymc3"][0])
        + list(CITATIONS["theano"][0])
    )
    bib = [
        CITATIONS["exoplanet"][1],
        CITATIONS["pymc3"][1],
        CITATIONS["theano"][1],
    ]
    for k, v in model.__citations__.items():
        cite += list(v[0])
        bib.append(v[1])

    txt = (
        r"This research made use of \textsf{{exoplanet}} "
        r"\citep{{exoplanet}} and its dependencies \citep{{{0}}}."
    )
    txt = txt.format(", ".join(sorted(cite)))
    txt = textwrap.wrap(txt, width=width)

    return "\n".join(txt), "\n".join(bib)


CITATIONS = {
    "exoplanet": (
        ("exoplanet:exoplanet",),
        r"""
@misc{exoplanet:exoplanet,
  author = {Daniel Foreman-Mackey and Ian Czekala and Eric Agol and
            Rodrigo Luger and Geert Barentsen and Tom Barclay},
   title = {dfm/exoplanet: exoplanet v0.2.0},
   month = aug,
    year = 2019,
     doi = {10.5281/zenodo.3359880},
     url = {https://doi.org/10.5281/zenodo.3359880}
}
""",
    ),
    "pymc3": (
        ("exoplanet:pymc3",),
        r"""
@article{exoplanet:pymc3,
    title={Probabilistic programming in Python using PyMC3},
   author={Salvatier, John and Wiecki, Thomas V and Fonnesbeck, Christopher},
  journal={PeerJ Computer Science},
   volume={2},
    pages={e55},
     year={2016},
publisher={PeerJ Inc.}
}
""",
    ),
    "theano": (
        ("exoplanet:theano",),
        r"""
@article{exoplanet:theano,
    title="{Theano: A {Python} framework for fast computation of mathematical
            expressions}",
   author={{Theano Development Team}},
  journal={arXiv e-prints},
   volume={abs/1605.02688},
     year=2016,
    month=may,
      url={http://arxiv.org/abs/1605.02688}
}
""",
    ),
    "astropy": (
        ("exoplanet:astropy13", "exoplanet:astropy18"),
        r"""
@article{exoplanet:astropy13,
   author = {{Astropy Collaboration} and {Robitaille}, T.~P. and {Tollerud},
             E.~J. and {Greenfield}, P. and {Droettboom}, M. and {Bray}, E. and
             {Aldcroft}, T. and {Davis}, M. and {Ginsburg}, A. and
             {Price-Whelan}, A.~M. and {Kerzendorf}, W.~E. and {Conley}, A. and
             {Crighton}, N. and {Barbary}, K. and {Muna}, D. and {Ferguson}, H.
             and {Grollier}, F. and {Parikh}, M.~M. and {Nair}, P.~H. and
             {Unther}, H.~M. and {Deil}, C. and {Woillez}, J. and {Conseil}, S.
             and {Kramer}, R. and {Turner}, J.~E.~H. and {Singer}, L. and
             {Fox}, R. and {Weaver}, B.~A. and {Zabalza}, V. and {Edwards},
             Z.~I. and {Azalee Bostroem}, K. and {Burke}, D.~J. and {Casey},
             A.~R. and {Crawford}, S.~M. and {Dencheva}, N. and {Ely}, J. and
             {Jenness}, T. and {Labrie}, K. and {Lim}, P.~L. and
             {Pierfederici}, F. and {Pontzen}, A. and {Ptak}, A. and {Refsdal},
             B. and {Servillat}, M. and {Streicher}, O.},
    title = "{Astropy: A community Python package for astronomy}",
  journal = {\aap},
     year = 2013,
    month = oct,
   volume = 558,
    pages = {A33},
      doi = {10.1051/0004-6361/201322068},
   adsurl = {http://adsabs.harvard.edu/abs/2013A%26A...558A..33A},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}

@article{exoplanet:astropy18,
   author = {{Astropy Collaboration} and {Price-Whelan}, A.~M. and
             {Sip{\H o}cz}, B.~M. and {G{\"u}nther}, H.~M. and {Lim}, P.~L. and
             {Crawford}, S.~M. and {Conseil}, S. and {Shupe}, D.~L. and
             {Craig}, M.~W. and {Dencheva}, N. and {Ginsburg}, A. and
             {VanderPlas}, J.~T. and {Bradley}, L.~D. and
             {P{\'e}rez-Su{\'a}rez}, D. and {de Val-Borro}, M.
             and {Aldcroft}, T.~L. and {Cruz}, K.~L. and {Robitaille}, T.~P.
             and {Tollerud}, E.~J. and {Ardelean}, C. and {Babej}, T. and
             {Bach}, Y.~P. and {Bachetti}, M. and {Bakanov}, A.~V. and
             {Bamford}, S.~P. and {Barentsen}, G. and {Barmby}, P. and
             {Baumbach}, A. and {Berry}, K.~L.  and {Biscani}, F. and
             {Boquien}, M. and {Bostroem}, K.~A. and {Bouma}, L.~G. and
             {Brammer}, G.~B. and {Bray}, E.~M. and {Breytenbach}, H. and
             {Buddelmeijer}, H. and {Burke}, D.~J. and {Calderone}, G. and
             {Cano Rodr{\'{\i}}guez}, J.~L. and {Cara}, M. and {Cardoso},
             J.~V.~M. and {Cheedella}, S. and {Copin}, Y. and {Corrales}, L.
             and {Crichton}, D. and {D'Avella}, D. and {Deil}, C. and
             {Depagne}, {\'E}. and {Dietrich}, J.~P. and {Donath}, A. and
             {Droettboom}, M. and {Earl}, N. and {Erben}, T. and {Fabbro}, S.
             and {Ferreira}, L.~A. and {Finethy}, T. and {Fox}, R.~T. and
             {Garrison}, L.~H. and {Gibbons}, S.~L.~J. and {Goldstein}, D.~A.
             and {Gommers}, R. and {Greco}, J.~P. and {Greenfield}, P. and
             {Groener}, A.~M. and {Grollier}, F. and {Hagen}, A. and {Hirst},
             P. and {Homeier}, D. and {Horton}, A.~J. and {Hosseinzadeh}, G.
             and {Hu}, L. and {Hunkeler}, J.~S. and {Ivezi{\'c}}, {\v Z}. and
             {Jain}, A. and {Jenness}, T. and {Kanarek}, G. and {Kendrew}, S.
             and {Kern}, N.~S. and {Kerzendorf}, W.~E. and {Khvalko}, A. and
             {King}, J. and {Kirkby}, D. and {Kulkarni}, A.~M. and {Kumar}, A.
             and {Lee}, A.  and {Lenz}, D.  and {Littlefair}, S.~P. and {Ma},
             Z. and {Macleod}, D.~M. and {Mastropietro}, M. and {McCully}, C.
             and {Montagnac}, S. and {Morris}, B.~M. and {Mueller}, M. and
             {Mumford}, S.~J. and {Muna}, D. and {Murphy}, N.~A. and {Nelson},
             S. and {Nguyen}, G.~H. and {Ninan}, J.~P. and {N{\"o}the}, M. and
             {Ogaz}, S. and {Oh}, S. and {Parejko}, J.~K.  and {Parley}, N. and
             {Pascual}, S. and {Patil}, R. and {Patil}, A.~A.  and {Plunkett},
             A.~L. and {Prochaska}, J.~X. and {Rastogi}, T. and {Reddy Janga},
             V. and {Sabater}, J.  and {Sakurikar}, P. and {Seifert}, M. and
             {Sherbert}, L.~E. and {Sherwood-Taylor}, H. and {Shih}, A.~Y. and
             {Sick}, J. and {Silbiger}, M.~T. and {Singanamalla}, S. and
             {Singer}, L.~P. and {Sladen}, P.~H. and {Sooley}, K.~A. and
             {Sornarajah}, S. and {Streicher}, O. and {Teuben}, P. and
             {Thomas}, S.~W. and {Tremblay}, G.~R. and {Turner}, J.~E.~H. and
             {Terr{\'o}n}, V.  and {van Kerkwijk}, M.~H. and {de la Vega}, A.
             and {Watkins}, L.~L. and {Weaver}, B.~A. and {Whitmore}, J.~B. and
             {Woillez}, J.  and {Zabalza}, V. and {Astropy Contributors}},
    title = "{The Astropy Project: Building an Open-science Project and Status
              of the v2.0 Core Package}",
  journal = {\aj},
     year = 2018,
    month = sep,
   volume = 156,
    pages = {123},
      doi = {10.3847/1538-3881/aabc4f},
   adsurl = {http://adsabs.harvard.edu/abs/2018AJ....156..123A},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
""",
    ),
    "espinoza18": (
        ("exoplanet:espinoza18",),
        r"""
@article{exoplanet:espinoza18,
   author = {{Espinoza}, N.},
    title = "{Efficient Joint Sampling of Impact Parameters and Transit Depths
              in Transiting Exoplanet Light Curves}",
  journal = {Research Notes of the American Astronomical Society},
     year = 2018,
    month = nov,
   volume = 2,
   number = 4,
    pages = {209},
      doi = {10.3847/2515-5172/aaef38},
   adsurl = {http://adsabs.harvard.edu/abs/2018RNAAS...2d.209E},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
""",
    ),
    "kipping13": (
        ("exoplanet:kipping13",),
        r"""
@ARTICLE{exoplanet:kipping13,
   author = {{Kipping}, D.~M.},
    title = "{Efficient, uninformative sampling of limb darkening coefficients
              for two-parameter laws}",
  journal = {\mnras},
     year = 2013,
    month = nov,
   volume = 435,
    pages = {2152-2160},
      doi = {10.1093/mnras/stt1435},
   adsurl = {http://adsabs.harvard.edu/abs/2013MNRAS.435.2152K},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
""",
    ),
    "starry": (
        ("exoplanet:luger18",),
        r"""
@article{exoplanet:luger18,
   author = {{Luger}, R. and {Agol}, E. and {Foreman-Mackey}, D. and {Fleming},
             D.~P. and {Lustig-Yaeger}, J. and {Deitrick}, R.},
    title = "{starry: Analytic Occultation Light Curves}",
  journal = {\aj},
     year = 2019,
    month = feb,
   volume = 157,
    pages = {64},
      doi = {10.3847/1538-3881/aae8e5},
   adsurl = {http://adsabs.harvard.edu/abs/2019AJ....157...64L},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
""",
    ),
    "celerite": (
        ("exoplanet:foremanmackey17", "exoplanet:foremanmackey18"),
        r"""
@article{exoplanet:foremanmackey17,
   author = {{Foreman-Mackey}, D. and {Agol}, E. and {Ambikasaran}, S. and
             {Angus}, R.},
    title = "{Fast and Scalable Gaussian Process Modeling with Applications to
              Astronomical Time Series}",
  journal = {\aj},
     year = 2017,
    month = dec,
   volume = 154,
    pages = {220},
      doi = {10.3847/1538-3881/aa9332},
   adsurl = {http://adsabs.harvard.edu/abs/2017AJ....154..220F},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}

@article{exoplanet:foremanmackey18,
   author = {{Foreman-Mackey}, D.},
    title = "{Scalable Backpropagation for Gaussian Processes using Celerite}",
  journal = {Research Notes of the American Astronomical Society},
     year = 2018,
    month = feb,
   volume = 2,
   number = 1,
    pages = {31},
      doi = {10.3847/2515-5172/aaaf6c},
   adsurl = {http://adsabs.harvard.edu/abs/2018RNAAS...2a..31F},
  adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
""",
    ),
}



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np

import theano
import theano.tensor as tt
from theano.tests import unittest_tools as utt

from .integrated_limbdark import IntegratedLimbDarkOp


class TestIntegratedLimbDark(utt.InferShapeTester):
    def setUp(self):
        super(TestIntegratedLimbDark, self).setUp()
        self.op_class = IntegratedLimbDarkOp
        self.op = IntegratedLimbDarkOp(Nc=3)

    def get_args(self):
        np.random.seed(1234)

        c = tt.dvector()
        r = tt.dvector()
        t = tt.dvector()
        n = tt.dvector()
        aome2 = tt.dvector()
        e = tt.dvector()
        sinw = tt.dvector()
        cosw = tt.dvector()
        sini = tt.dvector()
        cosi = tt.dvector()
        dt = tt.dvector()
        f = theano.function(
            [c, r, t, dt, n, aome2, sini, cosi, e, sinw, cosw],
            self.op(c, r, t, dt, n, aome2, sini, cosi, e, sinw, cosw)[0],
        )

        c_val = np.array([0.14691226, 0.25709645, -0.01836403])
        r_val = np.random.uniform(0.01, 0.2, 1000)
        P_val = 35.0
        t_val = np.linspace(-0.5 * P_val, 0.5 * P_val, len(r_val))
        n_val = 2 * np.pi / P_val + np.zeros_like(r_val)
        e_val = np.linspace(0, 1.0, len(r_val) + 1)[:-1]
        aome2_val = 10 * (1 - e_val ** 2)
        w_val = np.linspace(-np.pi, np.pi, len(r_val))
        sinw_val = np.sin(w_val)
        cosw_val = np.cos(w_val)
        i_val = 0.5 * np.pi + np.linspace(-0.01, 0.01, len(r_val))
        sini_val = np.sin(i_val)
        cosi_val = np.cos(i_val)

        dt_val = np.random.uniform(0.01, 0.1, len(r_val))

        return (
            f,
            [c, r, t, dt, n, aome2, sini, cosi, e, sinw, cosw],
            [
                c_val,
                r_val,
                t_val,
                dt_val,
                n_val,
                aome2_val,
                sini_val,
                cosi_val,
                e_val,
                sinw_val,
                cosw_val,
            ],
        )

    def test_basic(self):
        f, _, in_args = self.get_args()
        out = f(*in_args)
        utt.assert_allclose(0.0, out[0])
        utt.assert_allclose(0.0, out[-1])

    def test_infer_shape(self):
        f, args, arg_vals = self.get_args()
        self._compile_and_check(args, self.op(*args), arg_vals, self.op_class)

    def test_grad(self):
        _, args, in_args = self.get_args()

        # This is a hack because we don't compute gradients wrt texp
        func = lambda *args: self.op(  # NOQA
            *([args[0]] + [in_args[1]] + list(args[1:]))
        )[0]

        utt.verify_grad(func, [in_args[0]] + list(in_args[2:]), eps=1e-8)


class TestCircIntegratedLimbDark(utt.InferShapeTester):
    def setUp(self):
        super(TestCircIntegratedLimbDark, self).setUp()

        class _DummyClass(IntegratedLimbDarkOp):
            def __init__(self, *args, **kwargs):
                kwargs["circular"] = True
                super(_DummyClass, self).__init__(*args, **kwargs)

        self.op_class = _DummyClass
        self.op = _DummyClass(Nc=3)

    def get_args(self):
        np.random.seed(1234)

        c = tt.dvector()
        r = tt.dvector()
        t = tt.dvector()
        n = tt.dvector()
        aome2 = tt.dvector()
        sini = tt.dvector()
        cosi = tt.dvector()
        dt = tt.dvector()
        nothing1 = tt.dscalar()
        nothing2 = tt.dscalar()
        nothing3 = tt.dscalar()
        f = theano.function(
            [c, dt, t, r, n, aome2, sini, cosi, nothing1, nothing2, nothing3],
            self.op(
                c, dt, t, r, n, aome2, sini, cosi, nothing1, nothing2, nothing3
            )[0],
        )

        c_val = np.array([0.14691226, 0.25709645, -0.01836403])
        r_val = np.random.uniform(0.01, 0.2, 1000)
        P_val = 35.0
        t_val = np.linspace(-0.5 * P_val, 0.5 * P_val, len(r_val))
        n_val = 2 * np.pi / P_val + np.zeros_like(r_val)
        aome2_val = 10 + np.zeros_like(r_val)
        i_val = 0.5 * np.pi + np.linspace(-0.01, 0.01, len(r_val))
        sini_val = np.sin(i_val)
        cosi_val = np.cos(i_val)

        dt_val = np.random.uniform(0.01, 0.1, len(r_val))

        return (
            f,
            [c, dt, t, r, n, aome2, sini, cosi, nothing1, nothing2, nothing3],
            [
                c_val,
                dt_val,
                t_val,
                r_val,
                n_val,
                aome2_val,
                sini_val,
                cosi_val,
                0.0,
                0.0,
                0.0,
            ],
        )

    def test_basic(self):
        f, _, in_args = self.get_args()
        out = f(*in_args)
        utt.assert_allclose(0.0, out[0])
        utt.assert_allclose(0.0, out[-1])

    def test_infer_shape(self):
        f, args, arg_vals = self.get_args()
        self._compile_and_check(args, self.op(*args), arg_vals, self.op_class)

    def test_grad(self):
        _, args, in_args = self.get_args()
        func = lambda *args: self.op(  # NOQA
            *([args[0]] + [in_args[1]] + list(args[1:]) + [0.0, 0.0, 0.0])
        )[0]
        utt.verify_grad(func, [in_args[0]] + list(in_args[2:-3]), eps=1e-8)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np

import theano
import theano.tensor as tt
from theano.tests import unittest_tools as utt

from .kepler import KeplerOp


class TestKeplerSolver(utt.InferShapeTester):
    def setUp(self):
        super(TestKeplerSolver, self).setUp()
        self.op_class = KeplerOp
        self.op = KeplerOp()

    def _get_M_and_f(self, e, E):
        M = E - e * np.sin(E)
        f = 2 * np.arctan(np.sqrt((1 + e) / (1 - e)) * np.tan(0.5 * E))
        return M, f

    def test_edge(self):
        E = np.array([0.0, 2 * np.pi, -226.2, -170.4])
        e = (1 - 1e-6) * np.ones_like(E)
        e[-1] = 0.9939879759519037
        M, f = self._get_M_and_f(e, E)

        M_t = tt.dvector()
        e_t = tt.dvector()
        func = theano.function([M_t, e_t], self.op(M_t, e_t))
        sinf0, cosf0 = func(M, e)

        assert np.all(np.isfinite(sinf0))
        assert np.all(np.isfinite(cosf0))
        utt.assert_allclose(np.sin(f), sinf0)
        utt.assert_allclose(np.cos(f), cosf0)

    def test_pi(self):
        e = np.linspace(0, 1.0, 100)
        E = np.pi + np.zeros_like(e)
        M, f = self._get_M_and_f(e, E)

        M_t = tt.dvector()
        e_t = tt.dvector()
        func = theano.function([M_t, e_t], self.op(M_t, e_t))
        sinf0, cosf0 = func(M, e)

        assert np.all(np.isfinite(sinf0))
        assert np.all(np.isfinite(cosf0))
        utt.assert_allclose(np.sin(f), sinf0)
        utt.assert_allclose(np.cos(f), cosf0)

    def test_solver(self):
        e = np.linspace(0, 1, 500)[:-1]
        E = np.linspace(-300, 300, 1001)
        e = e[None, :] + np.zeros((len(E), len(e)))
        E = E[:, None] + np.zeros_like(e)
        M, f = self._get_M_and_f(e, E)

        M_t = tt.matrix()
        e_t = tt.matrix()
        func = theano.function([M_t, e_t], self.op(M_t, e_t))
        sinf0, cosf0 = func(M, e)

        assert np.all(np.isfinite(sinf0))
        assert np.all(np.isfinite(cosf0))
        utt.assert_allclose(np.sin(f), sinf0)
        utt.assert_allclose(np.cos(f), cosf0)

    def test_infer_shape(self):
        np.random.seed(42)
        M = tt.dvector()
        e = tt.dvector()
        M_val = np.linspace(-10, 10, 50)
        e_val = np.random.uniform(0, 0.9, len(M_val))
        self._compile_and_check(
            [M, e], self.op(M, e), [M_val, e_val], self.op_class
        )

    def test_grad(self):
        np.random.seed(1234)
        M_val = np.concatenate(
            (
                np.linspace(-10, 10, 100),
                [
                    0.0,
                    -np.pi + 1e-3,
                    np.pi - 1e-3,
                    0.5 * np.pi,
                    -0.5 * np.pi,
                    1.5 * np.pi,
                    2 * np.pi + 1e-3,
                ],
            )
        )
        e_val = np.random.uniform(0, 0.9, len(M_val))

        a = lambda *args: tt.arctan2(*self.op(*args))  # NOQA
        utt.verify_grad(a, [M_val, e_val], eps=1e-8)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["SimpleTransitOrbit"]

import theano.tensor as tt


class SimpleTransitOrbit(object):
    """An orbit representing a set of planets transiting a common central

    This orbit is parameterized by the observables of a transiting system,
    period, phase, duration, and impact parameter.

    Args:
        period: The orbital period of the planets in days.
        t0: The midpoint time of a reference transit for each planet in days.
        b: The impact parameters of the orbits.
        duration: The durations of the transits in days.
        r_star: The radius of the star in ``R_sun``.

    """

    def __init__(self, period=None, t0=0.0, b=0.0, duration=None, r_star=1.0):
        self.period = tt.as_tensor_variable(period)
        self.t0 = tt.as_tensor_variable(t0)
        self.b = tt.as_tensor_variable(b)
        self.duration = tt.as_tensor_variable(duration)
        self.r_star = tt.as_tensor_variable(r_star)

        self._b_norm = self.b * self.r_star
        x2 = self.r_star ** 2 - self._b_norm ** 2
        self.speed = 2 * tt.sqrt(x2) / self.duration
        self._half_period = 0.5 * self.period
        self._ref_time = self.t0 - self._half_period

    def get_star_position(self, t):
        nothing = tt.zeros_like(t)
        return nothing, nothing, nothing

    def get_planet_position(self, t):
        return self.get_relative_position(t)

    def get_relative_position(self, t):
        """The planets' positions relative to the star

        Args:
            t: The times where the position should be evaluated.

        Returns:
            The components of the position vector at ``t`` in units of
            ``R_sun``.

        """
        dt = tt.mod(tt.shape_padright(t) - self._ref_time, self.period)
        dt -= self._half_period
        x = tt.squeeze(self.speed * dt)
        y = tt.squeeze(self._b_norm + tt.zeros_like(dt))
        z = -tt.ones_like(x)
        return x, y, z

    def get_planet_velocity(self, t):
        raise NotImplementedError("a SimpleTransitOrbit has no velocity")

    def get_star_velocity(self, t):
        raise NotImplementedError("a SimpleTransitOrbit has no velocity")

    def get_radial_velocity(self, t, output_units=None):
        raise NotImplementedError("a SimpleTransitOrbit has no velocity")

    def in_transit(self, t, r=None, texp=None):
        """Get a list of timestamps that are in transit

        Args:
            t (vector): A vector of timestamps to be evaluated.
            r (Optional): The radii of the planets.
            texp (Optional[float]): The exposure time.

        Returns:
            The indices of the timestamps that are in transit.

        """
        dt = tt.mod(tt.shape_padright(t) - self._ref_time, self.period)
        dt -= self._half_period
        if self.r is None:
            tol = 0.5 * self.duration
        else:
            x = (r + self.r_star) ** 2 - self._b_norm ** 2
            tol = tt.sqrt(x) / self.speed
        if texp is not None:
            tol += 0.5 * texp
        mask = tt.any(tt.abs_(dt) < tol, axis=-1)
        return tt.arange(t.size)[mask]



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["KeplerOp"]

import theano
from theano import gof
import theano.tensor as tt

from ..build_utils import get_cache_version, get_compile_args, get_header_dirs


class KeplerOp(gof.COp):
    __props__ = ()
    func_file = "./kepler.cc"
    func_name = "APPLY_SPECIFIC(kepler)"

    def __init__(self, **kwargs):
        super(KeplerOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self):
        return get_cache_version()

    def c_headers(self, compiler):
        return ["exoplanet/theano_helpers.h", "exoplanet/kepler.h"]

    def c_header_dirs(self, compiler):
        return get_header_dirs(eigen=False)

    def c_compile_args(self, compiler):
        return get_compile_args(compiler)

    def make_node(self, mean_anom, eccen):
        in_args = [
            tt.as_tensor_variable(mean_anom),
            tt.as_tensor_variable(eccen),
        ]
        return gof.Apply(self, in_args, [in_args[0].type(), in_args[0].type()])

    def infer_shape(self, node, shapes):
        return shapes[0], shapes[0]

    def grad(self, inputs, gradients):
        M, e = inputs
        sinf, cosf = self(M, e)

        bM = tt.zeros_like(M)
        be = tt.zeros_like(M)

        # e * cos(f)
        ecosf = e * cosf

        # 1 - e^2
        ome2 = 1 - e ** 2

        # Partials
        dfdM = (1 + ecosf) ** 2 / ome2 ** 1.5
        dfde = (2 + ecosf) * sinf / ome2

        if not isinstance(gradients[0].type, theano.gradient.DisconnectedType):
            bM += gradients[0] * cosf * dfdM
            be += gradients[0] * cosf * dfde

        if not isinstance(gradients[1].type, theano.gradient.DisconnectedType):
            bM -= gradients[1] * sinf * dfdM
            be -= gradients[1] * sinf * dfde

        return [bM, be]

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["TTVOrbit"]

import theano.tensor as tt

from .keplerian import KeplerianOrbit


class TTVOrbit(KeplerianOrbit):
    """A generalization of a Keplerian orbit with transit timing variations

    Only one of the arguments ``ttvs`` or ``transit_times`` can be given and
    the other will be computed from the one that was provided.

    Args:
        ttvs: A list (with on entry for each planet) of "O-C" vectors for each
            transit of each planet in units of days. "O-C" means the
            difference between the observed transit time and the transit time
            expected for a regular periodic orbit.
        transit_times: A list (with on entry for each planet) of transit times
            for each transit of each planet in units of days. These times will
            be used to compute the implied (least squares) ``period`` and
            ``t0`` so these parameters cannot also be given.

    """

    def __init__(self, *args, **kwargs):
        ttvs = kwargs.pop("ttvs", None)
        transit_times = kwargs.pop("transit_times", None)
        if ttvs is None and transit_times is None:
            raise ValueError(
                "one of 'ttvs' or 'transit_times' must be " "defined"
            )
        if ttvs is not None:
            self.ttvs = [tt.as_tensor_variable(ttv) for ttv in ttvs]
        else:
            if kwargs.pop("period", None) is not None:
                raise ValueError(
                    "a period cannot be given if 'transit_times' " "is defined"
                )

            self.transit_times = []
            self.ttvs = []
            period = []
            t0 = []
            for i, times in enumerate(transit_times):
                times = tt.as_tensor_variable(times)

                N = times.shape[0]
                AT = tt.stack(
                    (tt.arange(N, dtype=times.dtype), tt.ones_like(times)),
                    axis=0,
                )
                A = tt.transpose(AT)
                ATA = tt.dot(AT, A)
                ATy = tt.dot(AT, times)
                w = tt.slinalg.solve_symmetric(ATA, ATy)
                expect = tt.dot(w, AT)

                period.append(w[0])
                t0.append(w[1])
                self.ttvs.append(times - expect)
                self.transit_times.append(times)

            kwargs["period"] = tt.stack(period)
            kwargs["t0"] = tt.stack(t0)

        super(TTVOrbit, self).__init__(*args, **kwargs)
        self._base_time = 0.5 - self.t0 / self.period

        if ttvs is not None:
            self.transit_times = [
                self.t0[i] + self.period[i] * tt.arange(ttv.shape[0]) + ttv
                for i, ttv in enumerate(self.ttvs)
            ]

    def _warp_times(self, t):
        delta = tt.shape_padleft(t) / tt.shape_padright(self.period, t.ndim)
        delta += tt.shape_padright(self._base_time, t.ndim)
        ind = tt.cast(tt.floor(delta), "int64")
        dt = tt.stack(
            [
                ttv[tt.clip(ind[i], 0, ttv.shape[0] - 1)]
                for i, ttv in enumerate(self.ttvs)
            ],
            -1,
        )
        return tt.shape_padright(t) + dt



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "estimate_semi_amplitude",
    "estimate_minimum_mass",
    "lomb_scargle_estimator",
    "autocorr_estimator",
]

import numpy as np

try:
    from scipy.ndimage import gaussian_filter
except ImportError:
    gaussian_filter = None

import astropy.units as u
from astropy.stats import LombScargle


def _get_design_matrix(periods, t0s, x):
    if t0s is not None:
        return np.vstack(
            [
                np.cos(
                    2 * np.pi * (x - (t0s[i] - 0.25 * periods[i])) / periods[i]
                )
                for i in range(len(periods))
            ]
            + [np.ones(len(x))]
        ).T
    return np.concatenate(
        [
            (
                np.sin(2 * np.pi * x / periods[i]),
                np.cos(2 * np.pi * x / periods[i]),
            )
            for i in range(len(periods))
        ]
        + [np.ones((1, len(x)))],
        axis=0,
    ).T


def estimate_semi_amplitude(periods, x, y, yerr=None, t0s=None):
    """Estimate the RV semi-amplitudes for planets in an RV series

    Args:
        periods: The periods of the planets. Assumed to be in ``days`` if not
            an AstroPy Quantity.
        x: The observation times. Assumed to be in ``days`` if not an AstroPy
            Quantity.
        y: The radial velocities. Assumed to be in ``m/s`` if not an AstroPy
            Quantity.
        yerr (Optional): The uncertainty on ``y``.
        t0s (Optional): The time of a reference transit for each planet, if
            known.

    Returns:
        An estimate of the semi-amplitude of each planet in units of ``m/s``.

    """
    if yerr is None:
        ivar = np.ones_like(y)
    else:
        ivar = 1.0 / yerr ** 2

    periods = u.Quantity(np.atleast_1d(periods), unit=u.day)
    if t0s is not None:
        t0s = u.Quantity(np.atleast_1d(t0s), unit=u.day).value
    x = u.Quantity(np.atleast_1d(x), unit=u.day)
    y = u.Quantity(np.atleast_1d(y), unit=u.m / u.s)
    ivar = u.Quantity(np.atleast_1d(ivar), unit=(u.s / u.m) ** 2)

    D = _get_design_matrix(periods.value, t0s, x.value)
    w = np.linalg.solve(
        np.dot(D.T, D * ivar.value[:, None]), np.dot(D.T, y.value * ivar.value)
    )
    if t0s is not None:
        K = w[:-1]
    else:
        w = w[:-1]
        K = np.sqrt(w[::2] ** 2 + w[1::2] ** 2)
    return K


def estimate_minimum_mass(periods, x, y, yerr=None, t0s=None, m_star=1):
    """Estimate the minimum mass(es) for planets in an RV series

    Args:
        periods: The periods of the planets. Assumed to be in ``days`` if not
            an AstroPy Quantity.
        x: The observation times. Assumed to be in ``days`` if not an AstroPy
            Quantity.
        y: The radial velocities. Assumed to be in ``m/s`` if not an AstroPy
            Quantity.
        yerr (Optional): The uncertainty on ``y``.
        t0s (Optional): The time of a reference transit for each planet, if
            known.
        m_star (Optional): The mass of the star. Assumed to be in ``M_sun``
            if not an AstroPy Quantity.

    Returns:
        An estimate of the minimum mass of each planet as an AstroPy Quantity
        with units of ``M_jupiter``.

    """
    periods = u.Quantity(np.atleast_1d(periods), unit=u.day)
    m_star = u.Quantity(m_star, unit=u.M_sun)
    K = estimate_semi_amplitude(periods, x, y, yerr=yerr, t0s=t0s)
    m_J = K / 28.4329 * m_star.value ** (2.0 / 3)
    m_J *= (periods.to(u.year)).value ** (1.0 / 3)
    return m_J * u.M_jupiter


def lomb_scargle_estimator(
    x,
    y,
    yerr=None,
    min_period=None,
    max_period=None,
    filter_period=None,
    max_peaks=2,
    **kwargs
):
    """Estimate period of a time series using the periodogram

    Args:
        x (ndarray[N]): The times of the observations
        y (ndarray[N]): The observations at times ``x``
        yerr (Optional[ndarray[N]]): The uncertainties on ``y``
        min_period (Optional[float]): The minimum period to consider
        max_period (Optional[float]): The maximum period to consider
        filter_period (Optional[float]): If given, use a high-pass filter to
            down-weight period longer than this
        max_peaks (Optional[int]): The maximum number of peaks to return
            (default: 2)

    Returns:
        A dictionary with the computed ``periodogram`` and the parameters for
        up to ``max_peaks`` peaks in the periodogram.

    """
    if min_period is not None:
        kwargs["maximum_frequency"] = 1.0 / min_period
    if max_period is not None:
        kwargs["minimum_frequency"] = 1.0 / max_period

    # Estimate the power spectrum
    model = LombScargle(x, y, yerr)
    freq, power = model.autopower(method="fast", normalization="psd", **kwargs)
    power /= len(x)
    power_est = np.array(power)

    # Filter long periods
    if filter_period is not None:
        freq0 = 1.0 / filter_period
        filt = 1.0 / np.sqrt(1 + (freq0 / freq) ** (2 * 3))
        power *= filt

    # Find and fit peaks
    peak_inds = (power[1:-1] > power[:-2]) & (power[1:-1] > power[2:])
    peak_inds = np.arange(1, len(power) - 1)[peak_inds]
    peak_inds = peak_inds[np.argsort(power[peak_inds])][::-1]
    peaks = []
    for i in peak_inds[:max_peaks]:
        A = np.vander(freq[i - 1 : i + 2], 3)
        w = np.linalg.solve(A, np.log(power[i - 1 : i + 2]))
        sigma2 = -0.5 / w[0]
        freq0 = w[1] * sigma2
        peaks.append(
            dict(
                log_power=w[2] + 0.5 * freq0 ** 2 / sigma2,
                period=1.0 / freq0,
                period_uncert=np.sqrt(sigma2 / freq0 ** 4),
            )
        )

    return dict(periodogram=(freq, power_est), peaks=peaks)


def next_pow_two(n):
    """Returns the next power of two greater than or equal to `n`"""
    i = 1
    while i < n:
        i = i << 1
    return i


def autocorr_function(x):
    """Estimate the normalized autocorrelation function of a 1-D series

    .. note:: This is from `emcee <https://github.com/dfm/emcee>`_.

    Args:
        x: The series as a 1-D numpy array.

    Returns:
        The autocorrelation function of the time series.

    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    acf /= acf[0]
    return acf


def autocorr_estimator(
    x,
    y,
    yerr=None,
    min_period=None,
    max_period=None,
    oversample=2.0,
    smooth=2.0,
    max_peaks=10,
):
    """Estimate the period of a time series using the autocorrelation function

    .. note:: The signal is interpolated onto a uniform grid in time so that
        the autocorrelation function can be computed.

    Args:
        x (ndarray[N]): The times of the observations
        y (ndarray[N]): The observations at times ``x``
        yerr (Optional[ndarray[N]]): The uncertainties on ``y``
        min_period (Optional[float]): The minimum period to consider
        max_period (Optional[float]): The maximum period to consider
        oversample (Optional[float]): When interpolating, oversample the times
            by this factor (default: 2.0)
        smooth (Optional[float]): Smooth the autocorrelation function by this
            factor times the minimum period (default: 2.0)
        max_peaks (Optional[int]): The maximum number of peaks to identify in
            the autocorrelation function (default: 10)

    Returns:
        A dictionary with the computed autocorrelation function and the
        estimated period. For compatibility with the
        :func:`lomb_scargle_estimator`, the period is returned as a list with
        the key ``peaks``.

    """
    if gaussian_filter is None:
        raise ImportError("scipy is required to use the autocorr estimator")

    if min_period is None:
        min_period = np.min(np.diff(x))
    if max_period is None:
        max_period = x.max() - x.min()

    # Interpolate onto an evenly spaced grid
    dx = np.min(np.diff(x)) / float(oversample)
    xx = np.arange(x.min(), x.max(), dx)
    yy = np.interp(xx, x, y)

    # Estimate the autocorrelation function
    tau = xx - x[0]
    acor = autocorr_function(yy)
    smooth = smooth * min_period
    acor = gaussian_filter(acor, smooth / dx)

    # Find the peaks
    peak_inds = (acor[1:-1] > acor[:-2]) & (acor[1:-1] > acor[2:])
    peak_inds = np.arange(1, len(acor) - 1)[peak_inds]
    peak_inds = peak_inds[tau[peak_inds] >= min_period]

    result = dict(autocorr=(tau, acor), peaks=[])

    # No peaks were found
    if len(peak_inds) == 0 or tau[peak_inds[0]] > max_period:
        return result

    # Only one peak was found
    if len(peak_inds) == 1:
        result["peaks"] = [
            dict(period=tau[peak_inds[0]], period_uncert=np.nan)
        ]
        return result

    # Check to see if second peak is higher
    if acor[peak_inds[1]] > acor[peak_inds[0]]:
        peak_inds = peak_inds[1:]

    # The first peak is larger than the maximum period
    if tau[peak_inds[0]] > max_period:
        return result

    result["peaks"] = [dict(period=tau[peak_inds[0]], period_uncert=np.nan)]
    return result
